# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import abstractmethod
import logging
import threading

from iotcs.client import StorageObject
from iotcs.exception import StateException
from iotcs.messaging.client.device.storage import StorageDispatcher


# com.oracle.iot.client.impl

# TODO: mention commonality when there is an enterprise impl.
#    """
#    StorageObjectBase is the common implementation between the
#    implementations of oracle.iot.client.impl.device.StorageObject,
#    namely com.oracle.iot.client.impl.device.StorageObjectImpl and
#    com.oracle.iot.client.impl.enterprise.StorageObjectImpl
#    """


class _SyncCallbackDispatcher(threading.Thread):
    """
    A Thread used to dispatch storage synchronization callbacks.

    See :class:`.SyncCallback`
    """

    __slots__ = (
        'callback',
        'events'
    )

    def __init__(self, events, callback):
        """
        Create a `_SyncCallbackDispatcher` instance.

        :param callback: the :class:`.SyncCallback` instance
        :param events: a list of :class:`.SyncEvent`
        """
        # TODO: Does this name have to be unique ?
        super(_SyncCallbackDispatcher, self).__init__(
            name="iotcs.client.SyncCallbackDispatcher",
            daemon=True)
        self.callback = callback
        self.events = events

    def run(self):
        """Execute the `SyncCallbackDispatcher` thread."""
        for event in self.events:
            self.callback.onSync(event)


class StorageObjectBase(StorageObject):
    """A base call for storage object implementations."""

    __slots__ = (
        'delegate',
        'syncEvents',
        'eventLock'
    )

    def __init__(self, delegate):
        """Create a `StorageObjectBase` instance."""
        super(StorageObjectBase, self).__init__(delegate)
        self.delegate = delegate
        self.syncEvents = []
        self.eventLock = threading.RLock()

    def sync(self):
        """
        Notify the library to sync content with the storage cloud.

        The direction of the sync, upload or download, depends on whether
        :meth:`inputPath <.setInputPath>` or the
        :meth:`outputPath <.setOutputPath>` have been set.
        This method does not start any uploads or downloads if
        :meth:`sycn status <.getSyncStatus>` is other than
        :attr:`NOT_IN_SYNC <.SyncStatus.NOT_IN_SYNC>`.

        This is a non-blocking call. The sync is performed on a separate
        thread on this :class:`.StorageObject`.

        If the input path cannot be read, or the output path cannot be
        written, an `OSError` is raised. Any I/O exceptions
        during the background sync are reported through the virtual device
        :class:`error callback <.AbstractVirtualDevice.ErrorCallback>`
        of the virtual device.

        :raises ValueError: if both input path and output path are `None`
        """
        if self.getSyncStatus() == StorageObject.SyncStatus.NOT_IN_SYNC:
            if self.getOutputPath() is None and self.getInputPath() is None:
                raise StateException("input path or output path must be set")
            self.setSyncStatus(StorageObject.SyncStatus.SYNC_PENDING)
            self.addSyncEvent(self.createSyncEvent())
        else:
            self.addSyncEvent(self.createSyncEvent())

    def setOnSync(self, callback):
        """
        Set a :class:`.StorageObject.SyncCallback`.

        The callback is invoked when the content referred to by this
        :class:`.StorageObject` is synchronized.

        :param callback: a :class:`.StorageObject.SyncCallback` to invoke when
            there is an error setting a value, if `None`, the existing
            `syncCallback` is removed.
        """
        super(StorageObjectBase, self).setOnSync(callback)
        with self.eventLock:
            self.syncEvents.clear()

    def getDelegate(self):
        """
        Return `delegate`.

        :return: `delegate`
        """
        return self.delegate

    def getSyncEvents(self):
        """
        Return `syncEvents`.

        :return: `syncEvents`
        """
        with self.eventLock:
            return self.syncEvents

    def addSyncEvent(self, syncEvent):
        """
        Add `syncEvent` to `syncEvents`.

        If `syncCallback` is `None` do nothing.
        """
        if self.getSyncCallback() is None:
            return
        status = self.getSyncStatus()
        if status == StorageObject.SyncStatus.SYNC_PENDING or \
                status == StorageObject.SyncStatus.NOT_IN_SYNC:
            with self.eventLock:
                self.syncEvents.append(syncEvent)
        elif status == StorageObject.SyncStatus.SYNC_FAILED or \
                status == StorageObject.SyncStatus.IN_SYNC:
            # Immediate Callback
            _SyncCallbackDispatcher(syncEvent, self.getSyncCallback()).start()

    @abstractmethod
    def handleStateChange(self):
        """Perform necessary tasks when the state changes."""
        pass

    @abstractmethod
    def createSyncEvent(self):
        """
        Create a :class:`.SyncEvent`.

        :return: a :class:`SyncEvent` instance
        """
        pass

    class SyncEventImpl(StorageObject.SyncEvent):
        """
        An event passed to the :meth:`.setOnSync` :class:`.SyncCallback`.

        This event occurs when an attribute value's content refers to
        a :class:`.StorageObject` and has been successfully
        synchronized, or has failed to be synchronized.
        """

        def __init__(self, source, virtualDevice, name):
            """Create a `SyncEventImpl` instance."""
            super(StorageObjectBase.SyncEventImpl, self).__init__()
            self.source = source
            self.virtualDevice = virtualDevice
            self.name = name

        def getVirtualDevice(self):
            """
            Return the virtual device that is the source of the event.

            :return: the virtual device, or `None` if :meth:`.sync` was
                called independently
            """
            return self.virtualDevice

        def getName(self):
            """
            Return the name of the attribute, action, or format.

            Return the name of the attribute, action, or format that
            this event is associated with.

            :return: the name, or `None` if :meth:`.sync` was called
                independently
            """
            return self.name

        def getSource(self):
            """
            Return the :class:`.StorageObject` source of this event.

            :return the storageObject, never `None`
            """
            return self.source

# com.oracle.iot.client.impl.device


class StorageObjectImpl(StorageObjectBase):
    """An implementation of :class:`.StorageObjectBase`."""

    class _ProgressCallback(StorageDispatcher.ProgressCallback):
        """Implementation of :class:`.StorageDispatcher.ProgressCallback`."""

        def progress(self, progress):
            """
            Handle upload/download progress.

            :param progress: upload/download
                :class:`.StorageDispatcher.ProgressCallback`
            """
            delegate = progress.getStorageObject()
            storageObject = StorageObjectImpl._delegateMap[delegate]
            if storageObject is None:
                return
            oldStatus = storageObject.getSyncStatus()
            state = progress.getState()
            if state == StorageDispatcher.Progress.State.COMPLETED:
                storageObject.setSyncStatus(StorageObject.SyncStatus.IN_SYNC)
                del(StorageObjectImpl._delegateMap[delegate])
            elif state == StorageDispatcher.Progress.State.CANCELLED or \
                    state == StorageDispatcher.Progress.State.FAILED:
                storageObject.setSyncStatus(StorageObject.SyncStatus.SYNC_FAILED)  # noqa: E501
                del(StorageObjectImpl._delegateMap[delegate])
            elif state == StorageDispatcher.Progress.State.IN_PROGRESS or \
                    state == StorageDispatcher.Progress.State.INITIATED or \
                    state == StorageDispatcher.Progress.State.QUEUED:
                pass

            if oldStatus != storageObject.getSyncStatus():
                if storageObject.getInputPath() is not None:
                    storageObject.handleStateChange()
                syncCallback = storageObject.getSyncCallback()
                if syncCallback is not None:
                    _SyncCallbackDispatcher(
                        storageObject.getSyncEvents(),
                        storageObject.getSyncCallback()).start()

        def _getLogger(self):
            return logging.getLogger("iotcs.client.device.StorageObjectImpl._ProgressCallback")

    _progressCallback = _ProgressCallback()
    _delegateMap = dict()

    __slots__ = (
        #  The name of an attribute or a field
        'nameForSyncEvent',
        'deviceForSyncEvent',
        'storageDispatcher'
    )

    def __init__(self, directlyConnectedDevice, delegate):
        """
        Create a `StorageObjectImpl` instance.

        :param directlyConnectedDevice:
            the :class:`messaging.client.device.DirectlyConnectedDevice`
            with :class:`.messaging.client.StorageObject` attributes.
        :param delegate: the :class:`.messaging.client.StorageObject` to handle
            storage operations.
        """
        super(StorageObjectImpl, self).__init__(delegate)
        self.nameForSyncEvent = None
        self.deviceForSyncEvent = None
        self.storageDispatcher = \
            StorageDispatcher.getStorageDispatcher(directlyConnectedDevice)

    def sync(self):
        """Synchronize the content with storage cloud service."""
        if self.getSyncStatus() == StorageObject.SyncStatus.NOT_IN_SYNC:
            super(StorageObjectImpl, self).sync()
            if not self.storageDispatcher.isClosed():
                delegate = self.getDelegate()
                StorageObjectImpl._delegateMap[delegate] = self
                self.storageDispatcher.setProgressCallback(
                    StorageObjectImpl._progressCallback)
                self.storageDispatcher.queue(delegate)
        else:
            super(StorageObjectImpl, self).sync()

    def handleStateChange(self):
        """Handle changes in sync status."""
        if self.deviceForSyncEvent is not None:
            self.deviceForSyncEvent.handleStorageObjectStateChange(self)

    def createSyncEvent(self):
        """
        Create a :class:`.SyncEvent`.

        :return: a :class:`SyncEvent` instance
        """
        return StorageObjectBase.SyncEventImpl(self, self.deviceForSyncEvent,
                                               self.nameForSyncEvent)

    def setSyncEventInfo(self, virtualDevice, nameForSyncEvent):
        """
        Set the virtual device and name for a sync event.

        :param virtualDevice: a :class:`.VirtualDevice` instance
        :param nameForSyncEvent: a name for a sync event
        :return:
        """
        self.deviceForSyncEvent = virtualDevice
        self.nameForSyncEvent = nameForSyncEvent

    def __str__(self):
        nfse = self.nameForSyncEvent \
            if self.nameForSyncEvent is not None \
            else "None"
        dfse = str(self.deviceForSyncEvent) \
            if self.deviceForSyncEvent is not None \
            else "None"
        return f'[nameForSyncEvent={nfse},\ndeviceForSyncEvent={dfse}]'

