# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.


from abc import ABC
from abc import abstractmethod
from enum import IntEnum
import configparser
import datetime
import json
import os
import platform
import sys
import time

from importlib.util import find_spec


class Closeable(ABC):
    """
    A Closeable is a source or destination of data that can be closed. The
    close method is invoked to release resources that the object is
    holding (such as open files).
    """

    __slots__ = ()

    @abstractmethod
    def close(self):
        """
        Closes this instance releasing any system resources associated with
        it. If the instance is already closed then invoking this method has
        no effect.
        :raises IOException: if there is an error closing the instance
        """
        pass


# TODO: There should be an Enum for Filter operators rather than the String
# TODO: represencations '$and' etc.

class Filter(ABC):
    """
    A Filter that can be used for queries.
    The following literal operators are supported
    '$and', '$or', '$not', '$like', '$in', '$eq',
    '$lt', '$lte', '$gt', '$gte', '$ne', '$exists'

    *Example:*

    Devices activated in the last hour that implement a "com:acme:device"
    device model.

    The following code:

    ```
    f = Filter.and(
        Filter.gte(Device.Field.CREATED.alias(),
            (System.currentTimeMillis() - 3600000L)),
        Filter.eq(Device.Field.STATE.alias(), "ACTIVATED"),
        Filter.like(Device.Field.DEVICE_MODELS.alias() + ".urn",
            "urn:com:acme:device:%"));
    ```

    will create the following JSON object:

    ```
    {"$and":[
        {"created":{"$gte":1457137772894}},
        {"state":{"$eq":"ACTIVATED"}},
        {"deviceModels.urn":{"$like":"urn:com:acme:device:%"}}
    ]}
    """


    @classmethod
    def and_(cls, *filters):
        """
        Creates a query filter that performs the logical AND between sub filters provided as argument: (filter1 and filter2 and ...).
        
        :param filters: sub filters to combine
        :return: a new filter { "$and": [ filter1, filter2, .... ] }
        """
        return _CompositeFilter("$and", filters)


    @classmethod
    def or_(cls, *filters):
        """
        Creates a query filter that performs the logical OR between sub filters provided as argument: (filter1 or filter2 or ...).
        
        :param filters: sub filters to combine
        :return: a new filter  { "$or": [ filter1, filter2, .... ] }
        """
        return _CompositeFilter("$or", filters)


    @classmethod
    def not_(cls, filter):
        """
        Creates a query filter that performs the logical NOT on the filter provided as argument: (not filter).
        
        :param filter: on which to apply the NOT
        :return: a new filter { "$not": filter1 }
        """
        return _CompositeFilter("$not", None)


    @classmethod
    # TODO: is not valid in 3.6 @overloaded
    def eq(cls, op1, op2):
        """
        Creates a query filter that performs (op1 == op2).
        
        :param op1: first operand
        :param op2: second operand
        :return: a new filter { op1: { "$eq": op2 } }
        """
        return _SimpleFilter("$eq", op1, op2)

    #
    # @eq.register(object, str, int)
    # def eq_0(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 == op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$eq": op2 } }
    #     '''
    #     return SimpleFilter("$eq", op1, op2)
    #
    #
    # @eq.register(object, str, float)
    # def eq_1(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 == op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$eq": op2 } }
    #     '''
    #     return SimpleFilter("$eq", op1, op2)
    #
    #
    # @eq.register(object, str, bool)
    # def eq_2(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 == op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$eq": op2 } }
    #     '''
    #     return SimpleFilter("$eq", op1, op2)


    @classmethod
    # TODO: is this not valid in 3.6 @overloaded
    def gt(cls, op1, op2):
        """
        Creates a query filter that performs(op1 &gt; op2).
        
        :param op1: first operand
        :param op2: second operand
        :return: a new filter { op1: { "$gt": op2 } }
        """
        return _SimpleFilter("$gt", op1, op2)

    #
    # @gt.register(object, str, float)
    # def gt_0(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs(op1 &gt; op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$gt": op2 } }
    #     '''
    #     return SimpleFilter("$gt", op1, op2)


    @classmethod
    # TODO: is this not valid in 3.6 @overloaded
    def gte(cls, op1, op2):
        """
        Creates a query filter that performs(op1 &gt;= op2).
        
        :param op1: first operand
        :param op2: second operand
        :return: a new filter { op1: { "$gte": op2 } }
        """
        return _SimpleFilter("$gte", op1, op2)

    #
    # @gte.register(object, str, float)
    # def gte_0(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs(op1 &gt;= op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$gte": op2 } }
    #     '''
    #     return SimpleFilter("$gte", op1, op2)


    @classmethod
    # TODO: is this not valid in 3.6 @overloaded
    def lt(cls, op1, op2):
        """
        Creates a query filter that performs (op1 &lt; op2).
        
        :param op1: first operand
        :param op2: second operand
        :return: a new filter { op1: { "$lt": op2 } }
        """
        return _SimpleFilter("$lt", op1, op2)

    #
    # @lt.register(object, str, float)
    # def lt_0(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 &lt; op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$lt": op2 } }
    #     '''
    #     return SimpleFilter("$lt", op1, op2)


    @classmethod
    # TODO: is this not valid in 3.6 @overloaded
    def lte(cls, op1, op2):
        """
        Creates a query filter that performs(op1 &lt;= op2).
        
        :param op1: first operand
        :param op2: second operand
        :return: a new filter { op1: { "$lte": op2 } }
        """
        return _SimpleFilter("$lte", op1, op2)

    #
    # @lte.register(object, str, float)
    # def lte_0(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs(op1 &lt;= op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$lte": op2 } }
    #     '''
    #     return SimpleFilter("$lte", op1, op2)


    @classmethod
    # TODO: is this not valid in 3.6 @overloaded
    def ne(cls, op1, op2):
        """
        Creates a query filter that performs (op1 != op2).
        
        :param op1: first operand
        :param op2: second operand
        :return: a new filter { op1: { "$ne": op2 } }
        """
        return _SimpleFilter("$ne", op1, op2)

    #
    # @ne.register(object, str, long)
    # def ne_0(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 != op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$ne": op2 } }
    #     '''
    #     return SimpleFilter("$ne", op1, op2)
    #
    #
    # @ne.register(object, str, float)
    # def ne_1(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 != op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$ne": op2 } }
    #     '''
    #     return SimpleFilter("$ne", op1, op2)
    #
    #
    # @ne.register(object, str, bool)
    # def ne_2(cls, op1, op2):
    #     '''
    #     Creates a query filter that performs (op1 != op2).
    #
    #     :param op1: first operand
    #     :param op2: second operand
    #     :return: a new filter { op1: { "$ne": op2 } }
    #     '''
    #     return SimpleFilter("$ne", op1, op2)


    @classmethod
    # TODO: is this not valid in 3.6 @overloaded
    def in_(cls, key, values):
        """
        Creates a query filter that performs (key IN [value1, value2, ...])

        :param key:    the key to check
        :param values: the list of values to compare with
        :return: a new filter { key: { "$in" : [value1, value2, ...] }}
        """
        return _SimpleFilter("$in", key, values)

    #
    # @in_.register(object, str, long)
    # def in__0(cls, key, values):
    #     '''
    #     Creates a query filter that performs (key IN [value1, value2, ...])
    #
    #     :param key:    the key to check
    #     :param values: the list of values to compare with
    #     :return: a new filter { key: { "$in" : [value1, value2, ...] }}
    #     '''
    #     return SimpleFilter("$in", key, values)
    #
    #
    # @in_.register(object, str, float)
    # def in__1(cls, key, values):
    #     '''
    #     Creates a query filter that performs (key IN [value1, value2, ...])
    #
    #     :param key:    the key to check
    #     :param values: the list of values to compare with
    #     :return: a new filter { key: { "$in" : [value1, value2, ...] }}
    #     '''
    #     return SimpleFilter("$in", key, values)
    #
    #
    # @in_.register(object, str, bool)
    # def in__2(cls, key, values):
    #     '''
    #     Creates a query filter that performs (key IN [value1, value2, ...])
    #
    #     :param key:    the key to check
    #     :param values: the list of values to compare with
    #     :return: a new filter { key: { "$in" : [value1, value2, ...] }}
    #     '''
    #     return SimpleFilter("$in", key, values)


    @classmethod
    def exists(cls, op1, e):
        """
        Creates a query filter that checks if op1 exists
        
        :param op1: first operand
        :param e:   either {true} to check that exists, or {@code false} to check that not exists
        :return: a new filter performing { op1: { "$exists" : true/false }}
        """
        return _SimpleFilter("$exists", op1, e)


    @classmethod
    def like(cls, value, pattern):
        """
        Creates a query filter that checks if the value matches the pattern
        
        :param value:   the value to check
        :param pattern: the pattern to match
        :return: a new filter performing { op1: { "$like" : pattern }}
        """
        return _SimpleFilter("$like", value, pattern)

    @abstractmethod
    def toJson(self):
        """
        Return the JSONObject for this query filter.
        
        :return: a JSON representation of the filter.
        """
        pass

class _CompositeFilter(Filter):

    __slots__ = (
        '__operands',
        '__operator'
    )

    # TODO: Should operands be variadic '*' ?

    def __init__(self, operator, operands):
        """

        :param self:
        :param operator: one of `Filter` operators See `Filter`
        :param operands: a list of `Filter`
        :return:
        """
        super(_CompositeFilter, self).__init__()
        self.__operands = []
        self.__operator = operator
        if operands is None:
            self.__operands.append(None)
        else:
            for o in operands:
                self.__operands.append(o)

    def __str__(self):
        return json.dumps(self.toJson())

    # TODO: Need to standardize on what these "toJson" methods return
    # TODO: a String or a python object that represents JSON, there is
    # TODO: no JSON object per se

    def toJson(self):
        # Construct python objects and then encode to JSON

        if len(self.__operands) > 1:
            jsonArray = []
            for f in self.__operands:
                jsonArray.append(f.toJson())

            return { self.__operator : jsonArray}

        return {self.__operator : None if self.__operands[0] is None else self.__operands[0].toJson()}


class _SimpleFilter(Filter):
    __slots__ = (
        '__value'
    )
    # __value is a python object that will encoded to json

    # TODO: is this not valid in 3.6 @overloaded
    def __init__(self, operator, name, value):
        """

        :param operator:
        :param name:
        :param value:
        """
        super(_SimpleFilter, self).__init__()
        self.__value = { name : { operator : value}}

    def __str__(self):
        return json.dumps(self.toJson())

    def toJson(self):
        return self.__value

class Pair():
    """A convenience container to manage a single key and value."""

    __slots__ = (
        '__key',
        '__value',
    )

    def __init__(self, key, value):
        """
        Create a :class:`iotcs.client.impl.Pair`.

        :param key: the pair identifier
        :param value: the pair value
        """
        self.__key = key
        self.__value = value

    def getKey(self):
        """
        Return the key.

        :return: the key
        """
        return self.__key

    def getValue(self):
        """
        Return the value.

        :return: the value
        """
        return self.__value

class TimeManager:
    diffWithServerMilis = 0

    @staticmethod
    def currentTimeMillis(date=None):
        if date is None:
            current_time = int(round(time.time() * 1000))
            return  current_time + TimeManager.diffWithServerMilis
        else:
            return int(date.timestamp() * 1000)

    @staticmethod
    def setCurrentTimeMillis(serverTime):
        TimeManager.diffWithServerMilis = serverTime - TimeManager.currentTimeMillis()

    @staticmethod
    def getFormattedTime(timestamp=None):
        if timestamp is None:
            return time.strftime("%a %b %d %H:%M:%S %Z %Y")
        else:
            return datetime.datetime.fromtimestamp(timestamp/1000)\
                .strftime("%a %b %d %H:%M:%S %Z %Y")
    def dateFromMillis(timestamp):
        return datetime.datetime.fromtimestamp(timestamp / 1000)


# TODO: Make Module scope, not need for class, probably need more unique name.
class Util():
    """
    A utility class for the library.
    """
    _CONFIG_MODULE_ROOT = 'iotcs'
    _CONFIG_MODULE_FILE = os.path.join('config', 'config.ini')

    # TODO: Define these in the classes that own them
    CONFIG_DEVICE_MESSAGING_SECTION = "device_messaging"

    _defaultconfig = None

    @classmethod
    def getConfig(cls, modulePath=None, configFile=None):
        """
        Return the `configParser` dictionary for the library.
        This method Assumes there is a subdirectory of the module called
        `config` and contains a file called `config.ini`.
        The configparser used is configured for `ExtendedInterpolation`
        :param modulePath: the module path in `dot` notation. If `None` the
                top module of the library will be used `iotcs`
        :return: `None` if there is no `config` subdirectory or no 
                `confif.ini` file in the `config` directory
        :raises: any exception possible from the use of `ConfigParser`,
                `importlib.util.find_spec`, or `os.path` methods.
        """

        # cache the default config file parser
        if modulePath is None and configFile is None:
            if cls._defaultconfig is not None:
                return cls._defaultconfig

        # Get the module spec to get the base path of the config
        # directory or whatever the top of the class hierarchy is
        _modulePath = cls._CONFIG_MODULE_ROOT if modulePath is None else modulePath
        _configFile = cls._CONFIG_MODULE_FILE if configFile is None else configFile

        mspec = find_spec(_modulePath)
        # this does not return the sudirectories of the module, just the
        # module path.
        subdirs = mspec.submodule_search_locations
        config = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())

        for p in subdirs:
            configini = os.path.join(p, _configFile)
            if not os.path.exists(configini):
                continue
            # reads all <_CONFIG_MDOULE_ROOT>/<subdir>/<_CONFIG_MODULE_FILE>
            config.read(configini)

        if modulePath is None and configFile is None:
            cls._defaultconfig = config

        # len(config is never 0)
        return config # if len(config) != 0 else None

    # PROPERTY_OS_NAME = "OSName"
    # PROPERTY_OS_VERSION = "OSVersion"
    # DEFAULT_SECRET = 'secret'
    # PUBILIC_KEY_ENCODING_FORMAT_X509 = 'X.509'
    SECRET_HASH_ALGO = 'HmacSHA256'

    @classmethod
    def getOSName(cls):
        osname = sys.platform
        return osname

    @classmethod
    def getOSversion(cls):
        rel = platform.release()
        return rel

    @classmethod
    def getPropertyInt(property, defaultVal):
        val = None
        if (property in os.environ):
            propval = os.environ.get(property)
            if (isinstance(propval, int)):
                val = propval
        if (val is None):
            val = defaultVal
        return  val

    @classmethod
    def strtobool(cls, boolstr):
        # Support only None, True, true or False, false
        if boolstr is None:
            return False
        return str(True).lower() == boolstr.lower()


class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Also, the decorated class cannot be
    inherited from. Other than that, there are no restrictions that apply
    to the decorated class.

    To get the singleton instance, use the `Instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    """

    def __init__(self, decorated):
        self._decorated = decorated

    def getInstance(self):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.

        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)

class StatusCode(IntEnum):
    OK=200
    CREATED=201
    ACCEPTED=202
    NON_AUTHORITATIVE_INFORMATION=203
    NO_CONTENT=204
    FINISHED=205
    PARTIAL_CONTENT=206
    BAD_REQUEST=400
    UNAUTHORIZED=401
    PAYMENT_REQUIRED=402
    FORBIDDEN=403
    NOT_FOUND=404
    METHOD_NOT_ALLOWED=405
    NOT_ACCEPTABLE=406
    REQUEST_TIMEOUT=408
    CONFLICT=409
    PRECON_FAILED=412
    TOO_MANY_REQUESTS=429
    INTERNAL_SERVER_ERROR=500
    NOT_IMPLEMENTED=501
    BAD_GATEWAY=502
    SERVICE_UNAVAILABLE=503
    GATEWAY_TIMEOUT=504
    OTHER=-1

class RestApi:
    """
    RestApi provides the REST api root URL.

    The configuration property *use_webapi* defined in the *rest_api*
    section of :mod:`iotcs.config.config.py` is used to determine if the
    web REST api variants should be used for requests to the IoT server.
    If *use_webapi* is`True` then the request URL will use
    */iot/webapi/*, else */iot/api* will be used.
    """

    # TODO: Why are these needed, I thought they would be created dynamically
    V1 = None
    V2 = None

    __slots__ = (
        '__reqRoot',
        '__privateRoot',
        '__isWebApi'
    )

    def __init__(self, version):
        """
        Initialize the :class:`.RestApi` constants.

        :param version: the rest api version to identify the rest resource
        """
        config = Util.getConfig()
        self.__isWebApi = config["rest_api"].getboolean("use_webapi")
        self.__reqRoot = ("/iot/webapi/" if self.__isWebApi
                          else "/iot/api/") + version
        self.__privateRoot = ("/iot/privatewebapi/" if self.__isWebApi
                              else "/iot/privateapi/") + version

    def getReqRoot(self):
        """Return the REST api request root URL."""
        return self.__reqRoot

    def getPrivateRoot(self):
        """Return the private REST api request root URL."""
        return self.__privateRoot

    def isWebApi(self):
        """
        Return `True` if the web api REST resource URL must be used.

        If the `iotcs.config.config.py` *rest_api* section property
        *use_webapi* is `True` then the web api REST resource URL must be
        used to make requests to the IoT server.
        :return: `True` if the *use_webapi* configuration property is `True`
        else `False`
        """
        return self.__isWebApi


RestApi.V1 = RestApi("v1")
RestApi.V2 = RestApi("v2")
