# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and 
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.


from abc import abstractmethod
from enum import Enum
import logging
import threading

from iotcs.exception import ArgumentException
from iotcs.exception import StateException

from iotcs.messaging.client.storage import StorageConnectionBase

from iotcs.messaging.client.storage import \
    StorageObjectDelegate as ClientStorageObjectDelegate

from iotcs.common import Closeable

# com.oracle.iot.client.impl.device

class StorageConnectionImpl(StorageConnectionBase):
    """
    `StorageConnection` transfers content to the Storage Cloud Service.

    There is one `StorageConnection` instance per client.
    """
    def __init__(self, secureConnection):
        """Create a `StorageConnectionImpl` instance."""
        super(StorageConnectionImpl, self).__init__(secureConnection)

    def _createStorageObject(self, storageContainerUrl, name, contentType,
                            encoding, date, length):
        """
        Return a `StorageObject` instance.

                Return an appropriate storage object delegate.

        There are two implementations, one in messaging which returns a
        :class:`.messaging.client.StorageObjectDelegate` and one for enterprise
        that returns `.enterprise.client.StorageObjectDelegate`.

        See :class:`.messaging.client.device.StorageObjectDelegate`

        :param storageContainerUrl:
        :param name:
        :param contentType:
        :param encoding:
        :param date:
        :param length:
        :return: a :class:
        """
        return StorageObjectDelegate(self, storageContainerUrl, name,
                                     contentType, encoding, date, length)

# com.oracle.iot.client.impl.device

class StorageObjectDelegate(ClientStorageObjectDelegate):
    """
    `StorageObjectDelegate` to provides progress state.

    See :class:`.messaging.client.StorageObjectDelegate`
    See :class:`.messaging.client.StorageObject`
    See :class:`.StorageDispatcher.Progress.State`

    This is the device library version of this class. It is
    instantiated by
    :mod:`iotcs.messaging.client.device.StorageConnectionImpl`
    """

    __slots__ = (
        'state'
        # An instance of :class:`.StorageDispatcher.Progress.State`.
    )

    def __init__(self, storageConnection, uri, name, contentType,
                 contentEncoding, dateOfLastModification, length):
        """Create a `StorageObjectDelegate` instance."""
        super(StorageObjectDelegate, self).__init__(storageConnection,
                                                    uri, name, contentType,
                                                    contentEncoding,
                                                    dateOfLastModification,
                                                    length)
        self.state = None

    def setState(self, state):
        """
        Set the transfer progress state.

        :param state: an instance of
            :class:`.StorageDispatcher.Progress.State`
        """
        self.state = state

    def getState(self):
        """
        Return the transfer progress state.

        See :class:`.StorageDispatcher.Progress.State`"""

        return self.state

    PROGRESS_UPDATE_INTERVAL = 1024 * 1024
    """
    The number of bytes transferred between progress updates
    """


    def setTransferredBytes(self, transferredBytes):
        """
        Set the number of bytes transferred.

        :param transferredBytes: the number of byts transferred.
        """
        diff = transferredBytes - self.transferredBytes
        super(StorageObjectDelegate,
              self).setTransferredBytes(transferredBytes)
        if diff >= self.PROGRESS_UPDATE_INTERVAL:
            #  TODO: IOT-42148
            pass

    def setOutputStream(self, outputStream):
        """
        Set the download output :class:`stream <.Path>`

        :param outputStream: the download output stream
        """
        if outputStream is not None and self.state is not None and \
                self.state != StorageDispatcher.Progress.State.COMPLETED and \
                self.state != StorageDispatcher.Progress.State.CANCELLED and \
                self.state != StorageDispatcher.Progress.State.FAILED:
            raise StateException("sync in progress")
        super(StorageObjectDelegate, self).setOutputStream(outputStream)
        self.setState(None)

    def setInputStream(self, inputStream):
        """Set the upload input :class:`stream <.Path>`"""
        if inputStream != None and self.state is not None and \
                self.state != StorageDispatcher.Progress.State.COMPLETED and \
                self.state != StorageDispatcher.Progress.State.CANCELLED and \
                self.state != StorageDispatcher.Progress.State.FAILED:
            raise StateException("sync in progress")
        super(StorageObjectDelegate, self).setInputStream(inputStream)
        self.setState(None)

    def isCancelled(self):
        """Return `True` if the transfer has been cancelled else `False`. """
        return self.state == StorageDispatcher.Progress.State.CANCELLED

# com.oracle.iot.client.device.util


class StorageDispatcher(Closeable):
    """
    The StorageDispatcher queues content for automatic upload or download.

    The StorageDispatcher uploads/downloads
    :class:`.messaging.client.StorageObject` content to/from, the Oracle
    Storage Cloud Service.

    .. code:: python

        dcd = DirectlyConnectedDevice(assetsFilePath, assetsFilePassword)
        storageDispatcher = StorageDispatcher.getStorageDispatcher(dcd)
        class UploadProgressCallback(StorageDispatcher.ProgressCallback):
            def progress(self, progress):
                storageObject = progress.getStorageObject()
                if storageObject.getName() == "lenna":
                    if progress.getState() ==
                            StorageDispatcher.Progress.State.COMPLETED:
                    # Can now safely remove images/Lenna.jpg, if desired.
                    elif progress.getState() ==
                            StorageDispatcher.Progress.State.IN_PROGRESS and
                            takingTooLong:
                        storageDispatcher.cancel(storageObject)

        storageDispatcher.setProgressCallback(UploadProgressCallback())

        lenna =  dcd.createStorageObject("lenna", "image/jpeg")
        lenna.setInputStream(pathlib.Path("../images/Lenna.jpg"))
        storageDispatcher.queue(lenna)

        # Download example
        # :class:`.messaging.client.device.DirectlyConnectedDevice`
        dcd = DirectlyConnectedDevice(assetsFilePath, assetsFilePassword)
        storageDispatcher = StorageDispatcher.getStorageDispatcher(dcd)

        class DownloadProgressCallback(StorageDispatcher.ProgressCallback):
            def progress(self, progress):
                storageObject = progress.getStorageObject()
                if storageObject.getName() == "lenna":
                    if progress.getState ==
                            StorageDispatcher.Progress.State.COMPLETED:
                    # downloads/Lenna.jpg is ready
                    elif progress.getState() ==
                            StorageDispatcher.Progress.State.IN_PROGRESS and
                            takingTooLong:
                        storageDispatcher.cancel(storageObject)

        storageDispatcher.setProgressCallback(DownloadProgressCallback())

        lenna =  dcd.createStorageObject(uri)
        lenna.setOutputStream(pathlib.Path("../images/Lenna.jpg"))
        storageDispatcher.queue(lenna)

    """

    class Progress(object):
        """A class to convey progress via the :class:`.ProgressCallback`."""

        class State(Enum):
            """Upload/Download progress states."""

            INITIATED = u'INITIATED'
            """Initial state after upload() or download() is called."""
            QUEUED = u'QUEUED'
            """Up/download is queued and not yet started."""
            IN_PROGRESS = u'IN_PROGRESS'
            """Up/download is currently in progress."""
            COMPLETED = u'COMPLETED'
            """Up/download completed successfully."""
            CANCELLED = u'CANCELLED'
            """Up/download was cancelled before it completed."""
            FAILED = u'FAILED'
            """Up/download failed without completing."""

        @abstractmethod
        def getStorageObject(self):
            """
            Return the :class:`.messaging.client.device.StorageObject`.

            Return the :class:`.messaging.client.device.StorageObject`
            that was queued for which this progress event pertains.
            :return: the :class:`.messaging.client.device.StorageObject`
            """
            pass

        @abstractmethod
        def getState(self):
            """
            Return the transfer :class:`.Progress.State`.

            :return the transfer state :class:`.Progress.State`
            """
            pass

        @abstractmethod
        def getFailureCause(self):
            """
            Return the exception that resulted in transfer failure.

            The exception is either an `OSError` or
            :class:`.SecurityException`.
            This is useful if the progress state is
            :attr:`.Progress.State.FAILED`

            :return the exception causing the failure
            """
            pass

        @abstractmethod
        def getBytesTransferred(self):
            """
            Return the number of bytes transferred.

            This number can be compared to the length of content obtained by
            calling
            :meth:`.StorageObject.getLength`.
            :return: the number of bytes transferred
            """
            pass

    class ProgressCallback(object):
        """A callback interface for monitoring progress of queued content."""

        @abstractmethod
        def progress(self, progress):
            """
            Notify of progress for content transfer.

            :param progress: progress data
            """
            pass

    @classmethod
    def getStorageDispatcher(cls, directlyConnectedDevice):
        """
        Return the :class:`.StorageDispatcher`.

        Return a :class:`.StorageDispatcher` for the given
        :class:`.messaging.client.device.DirectlyConnectedDevice`.

        :param directlyConnectedDevice the {@link DirectlyConnectedDevice}
        :return: a :class:`.StorageDispatcher`
        """
        return \
            StorageDispatcherImpl.getStorageDispatcher(directlyConnectedDevice)

    @abstractmethod
    def queue(self, storageObject):
        """
        Add a :class:`.messaging.client.StorageObject` to the queue.

        The :class:`.messaging.client.StorageObject` content is
        uploaded/downloaded to/from the Storage Cloud.

        :param storageObject: The :class:`.messaging.client.StorageObject`
            content :class:`.messageing.client.storageObject` to be queued
        :raises ArgumentException: if the storage object is `None`
        :raises StateException: if the storage object is already queued or
            in progress
        """
        pass

    @abstractmethod
    def cancel(self, storageObject):
        """
        Cancel the transfer of content to or from storage.

        This call has no effect if the transfer is completed, already
        cancelled, has failed, or the :class:`.messaging.client.StorageObject`
        is not queued.

        :param storageObject: the content
            :class:`.messaging.client.StorageObject` to be cancelled.
        """
        pass

    @abstractmethod
    def setProgressCallback(self, callback):
        """
        Set a callback to be notified as the transfer progresses.

        :param callback: the :class:`.StorageDispatcher.ProgressCallback`
            to invoke, if `None` the existing callback is removed
        """
        pass


class StorageDispatcherImpl(StorageDispatcher):
    """
    An implementation of :class:`.StorageDispatcher`.

    `StorageDispatcherImpl` queues content for automatic upload to
    or download from the Oracle Storage Cloud Service.
    """

    class ProgressCallbackDispatcher(threading.Thread):
        """A `Thread` to dispatch a :class:`.ProgressCallback`."""

        __slots__ = (
            'callback',
            'progress'
        )

        def __init__(self, progress, callback):
            """
            Create a `ProgressCallbackDispatcher` instance.

            :param progress: a :class:`.Progress` instance
            :param callback: a :class:`.ProgressCallbacl` instance
            """
            super(StorageDispatcherImpl.ProgressCallbackDispatcher, self).__init__(
                name="iotcs.messaging.client.device.ProgressCallbackDispatcher",
                daemon=True)
            self.callback = callback
            self.progress = progress

        def run(self):
            """Call the progress callback."""
            if self.callback is not None:
                self.callback.progress(self.progress)

    class ProgressImpl(StorageDispatcher.Progress):
        """An implementation of :class:`.StorageDispatcher.Progress`."""

        __slots__ = (
            'storageObject',
            'exception',
            'bytesTransferred',
            'state'
        )

        # com.oracle.iot.client.StorageObject, cast to
        #       com.oracle.iot.client.impl.device.StorageObjectDelegate

        def __init__(self, storageObject):
            """
            Create a :class:`.ProgressImpl` instance.

            :param storageObject: a
                :class:`.messaging.client.device.StorageObjectDelegate`
                instance
            """
            super(StorageDispatcherImpl.ProgressImpl, self).__init__()
            self.storageObject = storageObject
            self.state = storageObject.getState()
            self.exception = None
            self.bytesTransferred = 0

        def setFailureCause(self, e):
            """
            Set the exception that caused the failure.

            :param e: the `Exception` that caused the failure
            """
            self.exception = e

        def setBytesTransferred(self, bytesTransferred):
            """
            Set the number of bytes transferred.

            :param bytesTransferred: the number of bytes transferred
            """
            self.bytesTransferred = bytesTransferred

        def setState(self, state):
            """
            Set the progress state.

            :param state: one of the :class:`.ProgressState` states
            """
            self.state = state

        def getStorageObject(self):
            """
            Return the `StorageObject`.

            :return: the storage object
                :class:`.messaging.client.device.StorageObjectDelegate`
                instance
            """
            return self.storageObject

        def getState(self):
            """Return the :class:`.ProgressState`."""
            return self.state

        def getFailureCause(self):
            """Return the `Exception` causing the failure."""
            return self.exception

        def getBytesTransferred(self):
            """Return the number of bytes transferred."""
            return self.bytesTransferred

    class ContentTransmitter(threading.Thread):
        """A `Thread` that initiates pending uploads and downloads."""

        __slots__ = (
            'sd',
            'running'
        )

        def __init__(self, sd):
            """
            Create a :class:`.ContentTransmitter` instance.

            :param sd: a instance of :class:`.StorageDispatcherImpl`
            """
            super(StorageDispatcherImpl.ContentTransmitter, self).__init__(
                name="iotcs.messaging.client.device.StorageDispatcherImpl.ContentTransmitter",
                daemon=True)
            self.running = False
            self.sd = sd

        def run(self):
            """
            Run :class:`.ContentTransmitter`.

            If `running` is not `True` exit the thread.
            If a request to close is pending or the queue is empty
            exit the thread.
            Get the next :class:`.StorageObjectDelegate` from the queue
            and call :meth:`.transferAndCallback` with the storage
            object delegate.
            """
            # TODO: Can we rely totally on "running" and not have to
            # TODO: use "isClosed" ?
            self.running = True
            while self.running:

                with self.sd.contentLock:
                    if self.sd.isClosed() and len(self.sd.queue_) == 0:
                        self.running = False
                        continue
                storageObjectDelegate = None
                with self.sd.contentLock:
                    try:
                        while not self.sd.isClosed() and \
                                len(self.sd.queue_) == 0:
                            # TODO: Used timed wait
                            self.sd.contentQueued.wait()
                        storageObjectDelegate = self.sd.queue_.pop()
                    # TODO: What is the python equivalent of
                    # TODO: InterruptedException ?
                    except Exception as e:
                        self.__getLogger().warning(
                            f'StorageDispatcherImpl.ContentTransmitter.run(): '
                            f'caught exception {e}')

                if storageObjectDelegate is not None:
                    if self.sd.progressCallback is not None:
                        p = StorageDispatcherImpl.ProgressImpl(
                            storageObjectDelegate)
                        self.sd.dispatchProgressCallback(p)
                    self.sd.transferAndCallback(storageObjectDelegate)

        def stop(self):
            """Stop the :clas:`.ContentThread`."""
            self.running = False

        def __getLogger(self):
            return logging.getLogger('iotcs.messaging.client.device.storage.StorageDispatcherImpl.ContentTransmitter')

    _dispatcherMap = dict()
    _dispatcherMapLock = threading.RLock()

    __slots__ = (
        'progressCallback',
        'contentThread',
        'contentLock',
        'contentQueued',
        'closed',
        'requestClose',
        'closeLock',
        'queue_'
    )

    @classmethod
    def getStorageDispatcher(cls, directlyConnectedDevice):
        """
        Return a :class:`.StorageDispatcher`.

        The :class:`.StorageDispatcher` returned is mapped to the
        `directlyConnectedDevice`

        :param directlyConnectedDevice: A
            :class:`iotcs.messaging.client.device.DirectlyConnectedDevice`
            instance
        :return: the :class:`.StorageDispatcher` associated with this
            `directlyConnectedDevice`
        """
        storageDispatcher = None
        with cls._dispatcherMapLock:
            try:
                storageDispatcher = cls._dispatcherMap[directlyConnectedDevice]
            except KeyError as ke:
                # ignore, not in the map
                pass

        if storageDispatcher is None or storageDispatcher.isClosed():
            storageDispatcher = StorageDispatcherImpl()
            cls._dispatcherMap[directlyConnectedDevice] = storageDispatcher
        return storageDispatcher

    def __init__(self):
        """Create a :class:`.StorageDispatcherImpl` instance."""
        super(StorageDispatcherImpl, self).__init__()
        self.closed = False
        self.requestClose = False
        self.closeLock = threading.RLock()

        self.contentLock = threading.RLock()
        self.contentQueued = threading.Condition(lock=self.contentLock)

        self.queue_ = []
        self.progressCallback = None

        self.contentThread = StorageDispatcherImpl.ContentTransmitter(self)
        self.contentThread.start()

    # com.oracle.iot.client.StorageObject, cast to
    #       com.oracle.iot.client.impl.device.StorageObjectDelegate

    def queue(self, storageObject):
        """
        Add `storageObject` to the dispatch queue.

        :param storageObject: a
            :class:`.messaging.client.device.StorageObjectDelegate` instance
        """
        if storageObject is None:
            raise ArgumentException("StorageObject cannot be null.")
        # just for clarity
        storageObjectDelegate = storageObject
        state = storageObjectDelegate.getState()
        if state is not None:
            if state == StorageDispatcher.Progress.State.COMPLETED:
                return
            elif state == StorageDispatcher.Progress.State.IN_PROGRESS:
                raise StateException("Transfer is pending.")
            # elif state == StorageDispatcher.Progress.State.QUEUED:
            #    pass
            #  fall through
        with self.contentLock:
            self.queue_.append(storageObjectDelegate)
            storageObjectDelegate.setState(
                StorageDispatcher.Progress.State.QUEUED)
            if self.progressCallback is not None:
                p = StorageDispatcherImpl.ProgressImpl(storageObject)
                self.dispatchProgressCallback(p)
            self.contentQueued.notifyAll()

    # com.oracle.iot.client.StorageObject, cast to
    #       com.oracle.iot.client.impl.device.StorageObjectDelegate

    def cancel(self, storageObject):
        """
        Cancel the transfer of `storageObject`.

        :param storageObject: a
            :class:`.messaging.client.device.StorageObjectDelegate` instance
        """
        cancelled = False
        with self.contentLock:
            state = storageObject.getState()
            if state == StorageDispatcher.Progress.State.QUEUED:
                if storageObject in self.queue_:
                    cancelled = self.queue_.remove(storageObject)
            if cancelled or state == \
                    StorageDispatcher.Progress.State.IN_PROGRESS:
                storageObject.setState(
                    StorageDispatcher.Progress.State.CANCELLED)

        if cancelled and self.progressCallback is not None:
            p = StorageDispatcherImpl.ProgressImpl(storageObject)
            self.dispatchProgressCallback(p)

    def setProgressCallback(self, callback):
        """Set the :class:`.ProgressCallback` instance."""
        self.progressCallback = callback

    def close(self):
        """
        Close this `StorageDispatcherImpl` instance.

        When called the `contentThread` is stopped and this instance is
        removed from `_dispatcherMap`.
        """
        with self.closeLock:
            if not self.closed:
                self.requestClose = True
                try:
                    self.contentThread.stop()
                    # unblock the thread
                    self.contentQueued.notifyAll()
                    # wait until it terminates
                    self.contentThread.join()
                except Exception as e:
                    # TODO: log
                    pass

                with StorageDispatcherImpl._dispatcherMapLock:
                    rmlist = []
                    for dcd, sdisp in \
                            StorageDispatcherImpl._dispatcherMap.items():
                        if sdisp == self:
                            rmlist.append(dcd)

                    for dcd in rmlist:
                        del(StorageDispatcherImpl._dispatcherMap[dcd])

                self.closed = True

    def isClosed(self):
        """
        Return `True` if this `StorageDispatcherImpl` is closed.

        :return: `True` if this instance is closed
        """
        with self.closeLock:
            return self.requestClose

    def transferAndCallback(self, storageObjectDelegate):
        """
        Transfer the `storageObjectDelegate` to the storage service.

        Transfer the `storageObjectDelegate` to the storage service and
        dispatch the :class:`.ProgressCallback`

        :param storageObjectDelegate: a :class:`.StorageObjectDelegate`
            instance
        :return: `True` if the :class:`progress <.Progress>` state is
            :attr:`.Progress.State.COMPLETED` else `False`
        """
        progress = StorageDispatcherImpl.ProgressImpl(storageObjectDelegate)
        state = None
        try:
            #  Blocking!
            storageObjectDelegate.sync()
            state = StorageDispatcher.Progress.State.COMPLETED
            progress.setBytesTransferred(storageObjectDelegate.getLength())
        except Exception as e:
            if storageObjectDelegate.isCancelled():
                state = StorageDispatcher.Progress.State.CANCELLED
            else:
                state = StorageDispatcher.Progress.State.FAILED
                progress.setFailureCause(e)

        storageObjectDelegate.setState(state)
        progress.setState(state)
        self.dispatchProgressCallback(progress)
        return state == StorageDispatcher.Progress.State.COMPLETED

    def dispatchProgressCallback(self, progress):
        """
        Dispatch the :class:`.ProgressCallback` with `progress`.

        See :meth:`.setProgressCallback`

        :param progress: the storage object :class:`.Progress`
        """
        StorageDispatcherImpl.ProgressCallbackDispatcher(
            progress, self.progressCallback).start()
